#include "config.h"
#include <stdio.h>
#include <fcntl.h>

/* fcntl.h included in qnx_map.h */

extern int qnx_first_data, qnx_first_bss, qnx_last_data, qnx_last_bss;

#ifndef NOTDEF
int qnx_out_initial = 0;
	 
#ifndef QNX_ALLOCATION_SIZE
#define QNX_ALLOCATION_SIZE (564*1024)
#endif

static char qnx_initial_buffer[QNX_ALLOCATION_SIZE];
static char *qnx_current_brk = qnx_initial_buffer;
static char *qnx_end_brk = &qnx_initial_buffer[QNX_ALLOCATION_SIZE];
#else
extern char qnx_initial_buffer[];
extern char *qnx_current_brk;
extern char *qnx_end_brk;
extern int qnx_out_initial;
#endif

/*
This file dumps and restores the _DATA and _BSS sections of the program that
are used by the program.  Anything else is either constant data or belongs to
a library function and is not touched.

At the times save_data and restore_data are used, it is safe to just exit
when we get an error.
*/

static fatalsys(str)
char *str;
{
  perror(str);
  exit(1);
}

static fatalerr(str)
char *str;
{
  fputs(str, stderr);
  putc('\n', stderr);
  exit(1);
}

static safe_read(fd, buf, cnt)
int fd, cnt;
char *buf;
{
  int rtn=read(fd, buf, cnt);
  if (rtn != cnt) {
    if (rtn < 0)
      fatalsys("read");
    else
      fatalerr("short count in read");
  }
  return(rtn);
}

static safe_write(fd, buf, cnt)
int fd, cnt;
char *buf;
{
  int rtn=write(fd, buf, cnt);
  if (rtn != cnt) {
    if (rtn < 0)
      fatalsys("write");
    else
      fatalerr("short count in write");
  }
  return(rtn);
}

struct dump_hdr {
  char *dataloc;
  int datasize;
  char *bssloc;
  int bsssize;
};

save_data(file)
char *file;
{
  struct dump_hdr hdr;
  int fd, overshoot;
  if(qnx_out_initial)
    fatalerr(
"Out of initial allocation. Need more memory (QNX_ALLOCATION_SIZE).");
  if ((overshoot = qnx_end_brk - qnx_current_brk) > 1024)
	  printf ("You could reduce QNX_ALLOCATION_SIZE by %d bytes\n",
			  overshoot);
  fd=open(file, O_WRONLY|O_CREAT, 0666);
  if(fd < 0)
    fatalsys("open");
  hdr.dataloc = ((char *) &qnx_first_data);
  hdr.datasize = ((char *) &qnx_last_data) - ((char *) &qnx_first_data);
  hdr.bssloc = ((char *) &qnx_first_bss);
  hdr.bsssize = ((char *) &qnx_last_bss) - ((char *) &qnx_first_bss);
  safe_write(fd, (char*)&hdr, sizeof(hdr));
  safe_write(fd, hdr.dataloc, hdr.datasize);
  safe_write(fd, hdr.bssloc, hdr.bsssize);
  close(fd);
}

restore_data(file)
char *file;
{
  struct dump_hdr hdr;
  int fd=open(file, O_RDONLY);
  if(fd < 0)
    fatalsys("open");
  safe_read(fd, (char*)&hdr, sizeof(hdr));
  /* Consistancy checks */
  if (hdr.dataloc != ((char *) &qnx_first_data) ||
      hdr.datasize != ((char *) &qnx_last_data) - ((char *) &qnx_first_data) ||
      hdr.bssloc != ((char *) &qnx_first_bss) ||
      hdr.bsssize != ((char *) &qnx_last_bss) - ((char *) &qnx_first_bss))
    fatalerr("Inconsistancy between dump file and program");
  safe_read(fd, hdr.dataloc, hdr.datasize);
  safe_read(fd, hdr.bssloc, hdr.bsssize);
  qnx_out_initial = 1;
  close(fd);
}

/*
Due to the inability to create an executable image (it is undocumented),
and the possible alteration of library initialized data, we instead use
a file to save and restore those portions of the program belonging to
emacs itself.  For this to work, all pointers used as is from the dumped
data must point to static data.  This is achieved by making sys_malloc
allocate from a static array during the pure load phase.
*/

#undef sbrk

/* This size (556*1024) is hand-tuned to barely make it.  You might need
   to use the debugger to find a new size if things change in the .elc
   files that are loaded as part of the pure code. */

/* #include <malloc.h> done from s-qnx.h */
/*
Allocate memory by pointing within a static array, or calling the real malloc
if the array is full.  If we fail, return -1.
*/

char *
sys_sbrk (incr)
     int incr;
{
	if(qnx_out_initial)
	{
		return (sbrk (incr));
	}
	else
	{
		/* allocate out of our area */
		char *temp, *ptr;
		ptr = qnx_current_brk + incr;
		if(ptr <= qnx_end_brk)
		{
			temp = qnx_current_brk;
			qnx_current_brk = ptr;
			return temp;
		}
		else
		{
			qnx_out_initial = 1; /* We're out of initial allocation */
			return (sys_sbrk (incr));
		}
	}
}

char* _nmalloc (size_t nch)
{
	return (malloc(nch));
}

char* _fmalloc (size_t nch)
{
	return (malloc(nch));
}

char* _bmalloc (size_t nch)
{
	return (malloc(nch));
}

void _nfree (char* p)
{
	free (p);
}

void _ffree (char* p)
{
	free (p);
}

void _bfree (char* p)
{
	free (p);
}
